package com.github.polpetta.mezzotre.telegram.callbackquery;

import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.pengrad.telegrambot.model.Message;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.model.request.ParseMode;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.EditMessageText;
import com.pengrad.telegrambot.request.SendMessage;
import io.vavr.control.Try;
import java.nio.charset.StandardCharsets;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Singleton;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.util.StringBuilderWriter;
import org.slf4j.Logger;

/**
 * SelectLanguageTutorial event is related to support locale change for a particular chat when the
 * interaction with a new user starts
 *
 * @author Davide Polonio
 * @since 1.0
 */
@Singleton
public class SelectLanguageTutorial implements Processor {

  public static final String EVENT_NAME = "selectLanguageTutorial";
  private final Executor threadPool;
  private final LocalizedMessageFactory localizedMessageFactory;
  private final Logger log;
  private final UUIDGenerator uuidGenerator;

  /**
   * Additional fields that are related to {@code changeLanguage} event
   *
   * @author Davide Polonio
   * @since 1.0
   */
  public enum Field {
    NewLanguage("newLanguage");

    private final String name;

    Field(String name) {
      this.name = name;
    }

    public String getName() {
      return name;
    }
  }

  /**
   * Possible values for the additional {@link Field} of {@code changeLanguage} event
   *
   * @author Davide Polonio
   * @since 1.0
   */
  public enum Language {
    English("en-US"),
    Italian("it-IT");

    private final String locale;

    Language(String locale) {
      this.locale = locale;
    }

    public String getLocale() {
      return locale;
    }
  }

  @Inject
  public SelectLanguageTutorial(
      @Named("eventThreadPool") Executor threadPool,
      LocalizedMessageFactory localizedMessageFactory,
      Logger log,
      UUIDGenerator uuidGenerator) {
    this.threadPool = threadPool;
    this.localizedMessageFactory = localizedMessageFactory;
    this.log = log;
    this.uuidGenerator = uuidGenerator;
  }

  @Override
  public String getEventName() {
    return EVENT_NAME;
  }

  @Override
  public CompletableFuture<Optional<BaseRequest<?, ?>>> process(
      CallbackQueryContext callbackQueryContext, Update update) {
    return CompletableFuture.supplyAsync(
            () ->
                Optional.of(callbackQueryContext.getFields().getTelegramChatId())
                    .map(Double::longValue)
                    // If we're desperate, search in the message for the chat id
                    .or(
                        () ->
                            Optional.ofNullable(update.callbackQuery().message())
                                .map(Message::messageId)
                                .map(Long::valueOf))
                    .filter(chatId -> chatId != 0L && chatId != Long.MIN_VALUE)
                    .flatMap(chatId -> new QTgChat().id.eq(chatId).findOneOrEmpty())
                    .map(
                        tgChat -> {
                          tgChat.setLocale(
                              (String)
                                  callbackQueryContext
                                      .getFields()
                                      .getAdditionalProperties()
                                      .getOrDefault(
                                          Field.NewLanguage.getName(), tgChat.getLocale()));
                          tgChat.save();
                          log.trace(
                              "Locale for chat "
                                  + tgChat.getId()
                                  + " is now set in "
                                  + tgChat.getLocale());
                          return tgChat;
                        })
                    .orElseThrow(
                        () ->
                            new NoSuchElementException(
                                "Unable to find telegram chat "
                                    + Double.valueOf(
                                            callbackQueryContext.getFields().getTelegramChatId())
                                        .longValue()
                                    + " in the database")),
            threadPool)
        .thenApplyAsync(
            // If we are here then we're sure there is at least a chat associated with this callback
            tgChat -> {
              final String message =
                  Try.of(
                          () -> {
                            final Locale locale = Locale.forLanguageTag(tgChat.getLocale());
                            final ToolManager toolManager =
                                localizedMessageFactory.createVelocityToolManager(locale);
                            final VelocityContext velocityContext =
                                new VelocityContext(toolManager.createContext());

                            velocityContext.put("hasHelpBeenShown", tgChat.getHasHelpBeenShown());

                            final StringBuilder content = new StringBuilder();
                            final StringBuilderWriter stringBuilderWriter =
                                new StringBuilderWriter(content);

                            toolManager
                                .getVelocityEngine()
                                .mergeTemplate(
                                    "/template/callbackQuery/selectLanguageTutorial.vm",
                                    StandardCharsets.UTF_8.name(),
                                    velocityContext,
                                    stringBuilderWriter);

                            stringBuilderWriter.close();
                            return content.toString();
                          })
                      .get();

              log.trace("SelectLanguageTutorial event - message to send back: " + message);

              final String callBackGroupToDelete = callbackQueryContext.getEntryGroup();
              final int delete =
                  new QCallbackQueryContext().entryGroup.eq(callBackGroupToDelete).delete();
              log.trace(
                  "Deleted "
                      + delete
                      + " entries regarding callback group "
                      + callBackGroupToDelete);

              final Optional<Integer> messageId =
                  Optional.ofNullable(update.callbackQuery().message()).map(Message::messageId);
              BaseRequest<?, ?> baseRequest;
              Optional<InlineKeyboardMarkup> helpButton = Optional.empty();
              if (!tgChat.getHasHelpBeenShown()) {
                // Add a button to show all the possible commands
                final String showMeTutorialString =
                    localizedMessageFactory
                        .createResourceBundle(Locale.forLanguageTag(tgChat.getLocale()))
                        .getString("button.showMeTutorial");

                final CallbackQueryMetadata callbackQueryMetadata =
                    new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
                        .withEvent(ShowHelp.EVENT_NAME)
                        .withTelegramChatId(tgChat.getId())
                        .build();

                final CallbackQueryContext callbackData =
                    new CallbackQueryContext(
                        uuidGenerator.generateAsString(),
                        uuidGenerator.generateAsString(),
                        callbackQueryMetadata);
                callbackData.save();

                helpButton =
                    Optional.of(
                        new InlineKeyboardMarkup(
                            new InlineKeyboardButton(showMeTutorialString)
                                .callbackData(callbackData.getId())));
              }

              if (messageId.isPresent()) {
                log.trace("message id is present - editing old message");
                final EditMessageText editMessageText =
                    new EditMessageText(tgChat.getId(), messageId.get(), message)
                        .parseMode(ParseMode.Markdown);
                helpButton.ifPresent(editMessageText::replyMarkup);
                baseRequest = editMessageText;
              } else {
                log.trace("no message id - sending a new message");
                final SendMessage sendMessage =
                    new SendMessage(tgChat.getId(), message).parseMode(ParseMode.Markdown);
                helpButton.ifPresent(sendMessage::replyMarkup);
                baseRequest = sendMessage;
              }

              return Optional.of(baseRequest);
            },
            threadPool);
  }
}
