package com.github.polpetta.mezzotre.telegram.callbackquery;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import com.github.polpetta.mezzotre.helper.Loader;
import com.github.polpetta.mezzotre.helper.TestConfig;
import com.github.polpetta.mezzotre.i18n.LocalizedMessageFactory;
import com.github.polpetta.mezzotre.orm.model.CallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.TgChat;
import com.github.polpetta.mezzotre.orm.model.query.QCallbackQueryContext;
import com.github.polpetta.mezzotre.orm.model.query.QTgChat;
import com.github.polpetta.mezzotre.util.UUIDGenerator;
import com.github.polpetta.types.json.CallbackQueryMetadata;
import com.github.polpetta.types.json.ChatContext;
import com.google.gson.Gson;
import com.pengrad.telegrambot.model.Update;
import com.pengrad.telegrambot.model.request.InlineKeyboardButton;
import com.pengrad.telegrambot.model.request.InlineKeyboardMarkup;
import com.pengrad.telegrambot.request.BaseRequest;
import com.pengrad.telegrambot.request.EditMessageText;
import com.pengrad.telegrambot.request.SendMessage;
import io.ebean.Database;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;
import org.junit.jupiter.api.*;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.slf4j.LoggerFactory;
import org.testcontainers.containers.PostgreSQLContainer;
import org.testcontainers.junit.jupiter.Container;
import org.testcontainers.junit.jupiter.Testcontainers;

@Tag("slow")
@Tag("database")
@Tag("velocity")
@Testcontainers
class SelectLanguageTutorialIntegrationTest {

  private static Gson gson;

  @Container
  private final PostgreSQLContainer<?> postgresServer =
      new PostgreSQLContainer<>(TestConfig.POSTGRES_DOCKER_IMAGE);

  private Database database;
  private SelectLanguageTutorial selectLanguageTutorial;
  private UUIDGenerator fakeUUIDGenerator;

  @BeforeAll
  static void beforeAll() {
    gson = new Gson();
  }

  @BeforeEach
  void setUp() throws Exception {
    database =
        Loader.connectToDatabase(Loader.loadDefaultEbeanConfigWithPostgresSettings(postgresServer));

    fakeUUIDGenerator = mock(UUIDGenerator.class);

    selectLanguageTutorial =
        new SelectLanguageTutorial(
            Executors.newSingleThreadExecutor(),
            new LocalizedMessageFactory(Loader.defaultVelocityEngine()),
            LoggerFactory.getLogger(SelectLanguageTutorial.class),
            fakeUUIDGenerator);
  }

  private static Stream<Arguments> getTestLocales() {
    return Stream.of(
        Arguments.of(
            SelectLanguageTutorial.Language.Italian,
            "_*Procede a bere una pozione al cui suo interno si trova uno strano liquido"
                + " multicolore*_\n"
                + "\n"
                + "Grazie! Ora che ho bevuto questa posizione modificata di Parlare con animali che"
                + " ho trovato ieri al negozio di pozioni magiche la \"Cristalleria Fermentatrice\""
                + " posso parlare con te nel linguaggio che preferisci!\n"
                + "\n"
                + "Puoi sempre cambiare le preferenze della tua lingua scrivendo"
                + " /selectLanguageTutorial nella chat.\n"
                + "\n"
                + "Sembra tu non abbia ancora visto cosa posso fare! Per avere una lista completa"
                + " delle mie abilità, scrivi /help nella chat in qualsiasi momento!"
                + " Alternativamente, puoi premere il bottone qui sotto.\n",
            "en-US",
            "Mostrami cosa puoi fare!",
            false),
        Arguments.of(
            SelectLanguageTutorial.Language.English,
            "_*Proceeds to drink a potion with a strange, multicolor liquid*_\n"
                + "\n"
                + "Thanks! Now that I drank this modified potion of Speak with animals that I've"
                + " found at the \"Crystal Fermentary\" magic potion shop yesterday I can speak"
                + " with you in the language that you prefer!\n"
                + "\n"
                + "You can always change your language settings by typing /selectLanguageTutorial"
                + " in the chat.\n"
                + "\n"
                + "It seems you haven't checked out what I can do yet! To have a complete list of"
                + " my abilities, type /help in chat at any time! Alternatively, you can click the"
                + " button down below.\n",
            "it-IT",
            "Show me what you can do!",
            false),
        Arguments.of(
            SelectLanguageTutorial.Language.Italian,
            "_*Procede a bere una pozione al cui suo interno si trova uno strano liquido"
                + " multicolore*_\n"
                + "\n"
                + "Grazie! Ora che ho bevuto questa posizione modificata di Parlare con animali che"
                + " ho trovato ieri al negozio di pozioni magiche la \"Cristalleria Fermentatrice\""
                + " posso parlare con te nel linguaggio che preferisci!\n"
                + "\n"
                + "Puoi sempre cambiare le preferenze della tua lingua scrivendo"
                + " /selectLanguageTutorial nella chat.\n\n",
            "en-US",
            "Mostrami cosa puoi fare!",
            true),
        Arguments.of(
            SelectLanguageTutorial.Language.English,
            "_*Proceeds to drink a potion with a strange, multicolor liquid*_\n"
                + "\n"
                + "Thanks! Now that I drank this modified potion of Speak with animals that I've"
                + " found at the \"Crystal Fermentary\" magic potion shop yesterday I can speak"
                + " with you in the language that you prefer!\n"
                + "\n"
                + "You can always change your language settings by typing /selectLanguageTutorial"
                + " in the chat.\n\n",
            "it-IT",
            "Show me what you can do!",
            true));
  }

  @ParameterizedTest
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  @MethodSource("getTestLocales")
  void shouldProcessChangeLanguageToDesiredOneSendMessage(
      SelectLanguageTutorial.Language language,
      String expectedResult,
      String startingLocale,
      String buttonLocale,
      boolean hasHelpBeenShown)
      throws Exception {

    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67");

    final Update update =
        gson.fromJson(
            "{\n"
                + "\"update_id\":10000,\n"
                + "\"callback_query\":{\n"
                + "  \"id\": \"4382bfdwdsb323b2d9\",\n"
                + "  \"from\":{\n"
                + "     \"last_name\":\"Test Lastname\",\n"
                + "     \"type\": \"private\",\n"
                + "     \"id\":1111111,\n"
                + "     \"first_name\":\"Test Firstname\",\n"
                + "     \"username\":\"Testusername\"\n"
                + "  },\n"
                + "  \"data\": \"Data from button callback\",\n"
                + "  \"inline_message_id\": \"1234csdbsk4839\"\n"
                + "}\n"
                + "}",
            Update.class);

    final long tgChatId = 1111111L;
    final ChatContext chatContext = new ChatContext();
    final TgChat tgChat = new TgChat(tgChatId, chatContext, startingLocale, hasHelpBeenShown);
    tgChat.save();

    final CallbackQueryMetadata callbackQueryMetadata =
        new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
            .withEvent("selectLanguageTutorial")
            .withTelegramChatId(tgChatId)
            .withAdditionalProperty(
                SelectLanguageTutorial.Field.NewLanguage.getName(), language.getLocale())
            .build();
    final String entryGroupId = "2e67774a-e4e4-4369-a414-a7f8bfe74b80";
    final CallbackQueryContext changeLanguageCallbackQueryContext =
        new CallbackQueryContext(
            "c018108f-6612-4848-8fca-cf301460d4eb", entryGroupId, callbackQueryMetadata);
    changeLanguageCallbackQueryContext.save();

    final CompletableFuture<Optional<BaseRequest<?, ?>>> processFuture =
        selectLanguageTutorial.process(changeLanguageCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = processFuture.get();
    final SendMessage gotMessage = (SendMessage) gotResponseOpt.get();
    assertEquals(expectedResult, gotMessage.getParameters().get("text"));

    final InlineKeyboardButton[][] replyMarkups =
        ((InlineKeyboardMarkup)
                gotMessage.getParameters().getOrDefault("reply_markup", new InlineKeyboardMarkup()))
            .inlineKeyboard();
    final List<InlineKeyboardButton> keyboardButtons =
        Stream.of(replyMarkups).flatMap(Stream::of).toList();
    if (!hasHelpBeenShown) {
      assertEquals(1, keyboardButtons.size());
      assertEquals(buttonLocale, keyboardButtons.get(0).text());

      verify(fakeUUIDGenerator, times(2)).generateAsString();
      assertEquals(
          1, new QCallbackQueryContext().id.eq("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5").findCount());
    } else {
      assertEquals(0, keyboardButtons.size());
      assertEquals(0, new QCallbackQueryContext().findCount());
    }

    final TgChat retrievedTgChat = new QTgChat().id.eq(tgChatId).findOne();
    assertNotNull(retrievedTgChat);
    assertEquals(language.getLocale(), retrievedTgChat.getLocale());

    assertEquals(0, new QCallbackQueryContext().entryGroup.eq(entryGroupId).findCount());
  }

  @ParameterizedTest
  @Timeout(value = 1, unit = TimeUnit.MINUTES)
  @MethodSource("getTestLocales")
  void shouldProcessChangeLanguageToDesiredOneEditMessage(
      SelectLanguageTutorial.Language language,
      String expectedResult,
      String startingLocale,
      String buttonLocale,
      boolean hasHelpBeenShown)
      throws Exception {

    when(fakeUUIDGenerator.generateAsString())
        .thenReturn("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5")
        .thenReturn("16507fbd-9f28-48a8-9de1-3ea1c943af67");

    final Update update =
        gson.fromJson(
            "{\n"
                + "  \"update_id\": 10000,\n"
                + "  \"callback_query\": {\n"
                + "    \"id\": \"4382bfdwdsb323b2d9\",\n"
                + "    \"from\": {\n"
                + "      \"last_name\": \"Test Lastname\",\n"
                + "      \"type\": \"private\",\n"
                + "      \"id\": 1111111,\n"
                + "      \"first_name\": \"Test Firstname\",\n"
                + "      \"username\": \"Testusername\"\n"
                + "    },\n"
                + "    \"message\": {\n"
                + "      \"message_id\": 2631,\n"
                + "      \"from\": {\n"
                + "        \"id\": 244745330,\n"
                + "        \"is_bot\": true,\n"
                + "        \"first_name\": \"test\",\n"
                + "        \"username\": \"test\"\n"
                + "      },\n"
                + "      \"date\": 1680276288,\n"
                + "      \"chat\": {\n"
                + "        \"id\": 4772108,\n"
                + "        \"type\": \"private\",\n"
                + "        \"username\": \"userTest\",\n"
                + "        \"first_name\": \"First Name\",\n"
                + "        \"last_name\": \"Last Name\"\n"
                + "      },\n"
                + "      \"text\": \"Previous message content\",\n"
                + "      \"reply_markup\": {\n"
                + "        \"inline_keyboard\": [\n"
                + "          [\n"
                + "            {\n"
                + "              \"text\": \"English\",\n"
                + "              \"callback_data\": \"b345ea60-391a-40a4-96ac-8c36c4366498\"\n"
                + "            },\n"
                + "            {\n"
                + "              \"text\": \"Italian\",\n"
                + "              \"callback_data\": \"ab6487fa-b010-4eb6-a316-5b60f3b6bc1e\"\n"
                + "            }\n"
                + "          ]\n"
                + "        ]\n"
                + "      }\n"
                + "    },\n"
                + "    \"data\": \"Data from button callback\",\n"
                + "    \"inline_message_id\": \"1234csdbsk4839\"\n"
                + "  }\n"
                + "}",
            Update.class);

    final long tgChatId = 1111111L;
    final ChatContext chatContext = new ChatContext();
    final TgChat tgChat = new TgChat(tgChatId, chatContext, startingLocale, hasHelpBeenShown);
    tgChat.save();

    final CallbackQueryMetadata callbackQueryMetadata =
        new CallbackQueryMetadata.CallbackQueryMetadataBuilder()
            .withEvent("selectLanguageTutorial")
            .withTelegramChatId(tgChatId)
            .withAdditionalProperty(
                SelectLanguageTutorial.Field.NewLanguage.getName(), language.getLocale())
            .build();
    final String entryGroupId = "2e67774a-e4e4-4369-a414-a7f8bfe74b80";
    final CallbackQueryContext changeLanguageCallbackQueryContext =
        new CallbackQueryContext(
            "c018108f-6612-4848-8fca-cf301460d4eb", entryGroupId, callbackQueryMetadata);
    changeLanguageCallbackQueryContext.save();

    final CompletableFuture<Optional<BaseRequest<?, ?>>> processFuture =
        selectLanguageTutorial.process(changeLanguageCallbackQueryContext, update);
    final Optional<BaseRequest<?, ?>> gotResponseOpt = processFuture.get();
    final EditMessageText gotMessage = (EditMessageText) gotResponseOpt.get();
    assertEquals(expectedResult, gotMessage.getParameters().get("text"));

    final InlineKeyboardButton[][] replyMarkups =
        ((InlineKeyboardMarkup)
                gotMessage.getParameters().getOrDefault("reply_markup", new InlineKeyboardMarkup()))
            .inlineKeyboard();
    final List<InlineKeyboardButton> keyboardButtons =
        Stream.of(replyMarkups).flatMap(Stream::of).toList();
    if (!hasHelpBeenShown) {
      assertEquals(1, keyboardButtons.size());
      assertEquals(buttonLocale, keyboardButtons.get(0).text());

      verify(fakeUUIDGenerator, times(2)).generateAsString();
      assertEquals(
          1, new QCallbackQueryContext().id.eq("e86e6fa1-fdd4-4120-b85d-a5482db2e8b5").findCount());
    } else {
      assertEquals(0, keyboardButtons.size());
      assertEquals(0, new QCallbackQueryContext().findCount());
    }

    final TgChat retrievedTgChat = new QTgChat().id.eq(tgChatId).findOne();
    assertNotNull(retrievedTgChat);
    assertEquals(language.getLocale(), retrievedTgChat.getLocale());

    assertEquals(0, new QCallbackQueryContext().entryGroup.eq(entryGroupId).findCount());
  }
}
